from copy import deepcopy
from typing import Iterable, Optional, Tuple, Union

from tdm.abstract.datamodel import AbstractFact, AbstractTalismanSpan, FactMetadata, FactStatus, FactType

ConceptValueType = str  # TODO: describe concept type


class ConceptFact(AbstractFact[str]):
    def __init__(self, id_: Optional[str], status: FactStatus, type_id: str,
                 value: Optional[Union[str, Tuple[str, ...]]] = None,
                 mention: Optional[Iterable[AbstractTalismanSpan]] = None,
                 metadata: Optional[FactMetadata] = None):
        metadata = deepcopy(metadata)  # guarantee immutability
        super().__init__(id_, FactType.CONCEPT, status, type_id, value, mention, metadata)

    def with_changes(self: 'ConceptFact', *, status: FactStatus = None, type_id: str = None,
                     value: Union[str, Tuple[str, ...]] = None,
                     mention: Tuple[AbstractTalismanSpan, ...] = None,
                     metadata: FactMetadata = None) -> 'ConceptFact':
        return ConceptFact(
            self._id,
            status if status is not None else self._status,
            type_id if type_id is not None else self._type_id,
            value if value is not None else self._value,
            mention if mention is not None else self._mention,
            metadata if metadata is not None else self._metadata
        )



from copy import deepcopy
from typing import Iterable, Optional, Tuple, Union

from tdm.abstract.datamodel import AbstractFact, AbstractTalismanSpan, FactMetadata, FactStatus, FactType
from tdm.abstract.datamodel.fact import AbstractLinkValue
from tdm.datamodel.fact.concept import ConceptFact
from tdm.datamodel.fact.relation import RelationFact
from tdm.datamodel.fact.value import ValueFact


class PropertyLinkValue(AbstractLinkValue[Union[ConceptFact, RelationFact], ValueFact]):
    @classmethod
    def validate_slots(cls, source: Union[ConceptFact, RelationFact], target: ValueFact):
        if source.fact_type is not FactType.CONCEPT and source.fact_type is not FactType.RELATION:
            raise ValueError(f"inconsistent property link source type: {source.fact_type}")
        if target.fact_type is not FactType.VALUE:
            raise ValueError(f"inconsistent property link target type: {target.fact_type}")


class PropertyFact(AbstractFact[PropertyLinkValue]):
    def __init__(self, id_: Optional[str], status: FactStatus, type_id: str, value: PropertyLinkValue,
                 mention: Iterable[AbstractTalismanSpan] = None, metadata: Optional[FactMetadata] = None):
        if not isinstance(value, PropertyLinkValue):
            raise ValueError(f"illegal value type for property fact: {type(value)}")
        metadata = deepcopy(metadata)  # guarantee immutability
        super().__init__(id_, FactType.PROPERTY, status, type_id, value, mention, metadata)

    def with_changes(self: 'PropertyFact', *, status: FactStatus = None, type_id: str = None,
                     value: PropertyLinkValue = None,
                     mention: Tuple[AbstractTalismanSpan, ...] = None,
                     metadata: FactMetadata = None) -> 'PropertyFact':
        return PropertyFact(
            self._id,
            status if status is not None else self._status,
            type_id if type_id is not None else self._type_id,
            value if value is not None else self._value,
            mention if mention is not None else self._mention,
            metadata if metadata is not None else self._metadata
        )



from copy import deepcopy
from typing import Iterable, Optional, Tuple

from tdm.abstract.datamodel import AbstractFact, AbstractTalismanSpan, FactMetadata, FactStatus, FactType
from tdm.abstract.datamodel.fact import AbstractLinkValue
from tdm.datamodel.fact.concept import ConceptFact


class RelationLinkValue(AbstractLinkValue[ConceptFact, ConceptFact]):
    @classmethod
    def validate_slots(cls, source: ConceptFact, target: ConceptFact):
        if source.fact_type is not FactType.CONCEPT:
            raise ValueError(f"inconsistent relation link source type: {source.fact_type}")
        if target.fact_type is not FactType.CONCEPT:
            raise ValueError(f"inconsistent relation link target type: {target.fact_type}")


class RelationFact(AbstractFact[RelationLinkValue]):
    def __init__(self, id_: Optional[str], status: FactStatus, type_id: str, value: RelationLinkValue,
                 mention: Iterable[AbstractTalismanSpan] = None, metadata: Optional[FactMetadata] = None):
        if not isinstance(value, RelationLinkValue):
            raise ValueError(f"illegal value type for relation fact: {type(value)}")
        metadata = deepcopy(metadata)  # guarantee immutability
        super().__init__(id_, FactType.RELATION, status, type_id, value, mention, metadata)

    def with_changes(self: 'RelationFact', *, status: FactStatus = None, type_id: str = None, value: RelationLinkValue = None,
                     mention: Tuple[AbstractTalismanSpan, ...] = None,
                     metadata: FactMetadata = None) -> 'RelationFact':

        return RelationFact(
            self._id,
            status if status is not None else self._status,
            type_id if type_id is not None else self._type_id,
            value if value is not None else self._value,
            mention if mention is not None else self._mention,
            metadata if metadata is not None else self._metadata
        )



from tdm.abstract.datamodel import AbstractFact, AbstractTreeDocumentContent, FactType
from tdm.json_schema.fact.common import AbstractFactModel, SpanModel
from tdm.json_schema.fact.concept import ConceptFactModel
from tdm.json_schema.fact.property import PropertyFactModel
from tdm.json_schema.fact.relation import RelationFactModel
from tdm.json_schema.fact.value import ValueFactModel

_FACT_MODELS = {
    FactType.CONCEPT: ConceptFactModel,
    FactType.VALUE: ValueFactModel,
    FactType.PROPERTY: PropertyFactModel,
    FactType.RELATION: RelationFactModel
}


def build_fact_model(fact: AbstractFact, doc: AbstractTreeDocumentContent) -> AbstractFactModel:
    model_type = _FACT_MODELS[fact.fact_type]

    mention = tuple(SpanModel.build(span, doc) for span in fact.mention) if fact.mention is not None else None
    value = model_type.build_value(fact.value)
    return model_type.construct(id=fact.id, status=fact.status, type_id=fact.type_id, value=value, mention=mention,
                                metadata=fact.metadata)



__all__ = [
    'PropertyLinkValueModel', 'SpanModel',
    'ConceptFactModel', 'PropertyFactModel', 'RelationFactModel', 'ValueFactModel',
    'build_fact_model'
]

from .common import PropertyLinkValueModel, SpanModel
from .concept import ConceptFactModel
from .factory import build_fact_model
from .property import PropertyFactModel
from .relation import RelationFactModel
from .value import ValueFactModel



from abc import abstractmethod
from typing import Any, Callable, Dict, Optional, Tuple

from pydantic import BaseModel

from tdm.abstract.datamodel import AbstractFact, AbstractTalismanSpan, AbstractTreeDocumentContent, FactMetadata, \
    FactStatus, FactType
from tdm.abstract.datamodel.fact import AbstractLinkValue
from tdm.datamodel import TalismanSpan


class SpanModel(BaseModel):
    node_id: str
    value: str
    start: int
    end: int

    def to_span(self) -> TalismanSpan:
        return TalismanSpan(start_idx=self.start, end_idx=self.end, id_=self.node_id)

    @classmethod
    def build(cls, span: AbstractTalismanSpan, doc: AbstractTreeDocumentContent) -> 'SpanModel':
        return cls.construct(node_id=span.node_id, value=doc.text_for(span), start=span.start_idx, end=span.end_idx)


class PropertyLinkValueModel(BaseModel):
    property_id: Optional[str]
    from_fact: str
    to_fact: str

    @classmethod
    def build(cls, value: AbstractLinkValue) -> 'PropertyLinkValueModel':
        return cls.construct(property_id=value.property_id, from_fact=value.from_fact.id, to_fact=value.to_fact.id)


class AbstractFactModel(BaseModel):
    id: str
    fact_type: FactType
    status: FactStatus
    type_id: str
    value: Any
    mention: Optional[Tuple[SpanModel, ...]]
    metadata: Optional[FactMetadata]

    def to_fact_factory(self) -> Tuple[FactType, Callable[[Dict[str, AbstractFact]], AbstractFact]]:
        def create_fact(mapping: Dict[str, AbstractFact]) -> AbstractFact:
            mention = tuple(span_model.to_span() for span_model in self.mention) if self.mention is not None else None
            return self.fact_factory(self.id, self.status, self.type_id, self.to_value(mapping), mention, self.metadata)

        return self.fact_type, create_fact

    @abstractmethod
    def to_value(self, mapping: Dict[str, AbstractFact]) -> Any:
        pass

    @property
    @abstractmethod
    def fact_factory(self) -> Callable[[str, FactStatus, str, Any, Tuple[AbstractTalismanSpan, ...],
                                       FactMetadata], AbstractFact]:
        pass

    @classmethod
    @abstractmethod
    def build_value(cls, value: Any) -> Any:
        pass

    def __hash__(self) -> int:
        return hash((self.id, self.fact_type, self.status, self.type_id))



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

with open("VERSION", "r", encoding="utf-8") as f:
    version = f.read()

setup(
    name='talisman-dm',
    version=version,
    description='Talisman Document Model python implementation',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://gitlab.at.ispras.ru/talisman/core/talisman-tdm',
    author='ISPRAS MODIS NLP team',
    author_email='modis@ispras.ru',
    maintainer='Vladimir Mayorov',
    maintainer_email='vmayorov@ispras.ru',
    packages=find_packages(include=['tdm', 'tdm.*']),
    install_requires=['pydantic>=1.4', 'typing-extensions>=4.0.1', 'intervaltree>=3.0.2'],
    data_files=[('', ['VERSION'])],
    python_requires='>=3.6',
    license='Apache Software License',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: Apache Software License'
    ]
)



from typing import Optional

from pydantic import BaseModel

from tdm.abstract.datamodel import NodeType
from tdm.datamodel import TreeDocumentContent


class NodeMetadata(BaseModel):
    node_type: NodeType
    original_text: Optional[str]
    hidden: bool = False

    class Config:
        extra = 'allow'  # any other extra fields will be kept

    def parse(self) -> dict:
        return self.dict()

    @classmethod
    def build(cls, node: TreeDocumentContent) -> 'NodeMetadata':
        return cls.construct(
            node_type=node.type,
            original_text=node.original_node_text if node.original_node_text != node.node_text else None,
            hidden=node.is_hidden
        )



from abc import abstractmethod
from typing import Optional, Tuple, Type, TypeVar


from pydantic import BaseModel

from .datamodel import AbstractTreeDocumentContent

_AbstractContentModel = TypeVar('_AbstractContentModel', bound='AbstractContentModel')


class AbstractContentModel(BaseModel):
    @abstractmethod
    def to_content(self) -> AbstractTreeDocumentContent:
        pass

    @classmethod
    @abstractmethod
    def build(cls: Type[_AbstractContentModel], doc: AbstractTreeDocumentContent) -> _AbstractContentModel:
        pass


class DocumentMetadataFields(BaseModel):
    title: Optional[str]
    file_name: Optional[str]
    file_type: Optional[str]
    size: Optional[int]
    created_time: Optional[int]
    access_time: Optional[int]
    modified_time: Optional[int]
    publication_date: Optional[int]
    publication_author: Optional[str]
    description: Optional[str]
    parent_uuid: Optional[str]
    url: Optional[str]
    access_level: Optional[str]
    user: Optional[str]
    path: Optional[str]
    trust_level: Optional[float]
    markers: Optional[Tuple[str, ...]]
    related_concept_id: Optional[str]

    class Config:
        extra = 'allow'  # any other extra fields will be kept


class FactMetadataFields(BaseModel):
    created_time: Optional[int]
    modified_time: Optional[int]

    class Config:
        extra = 'allow'  # any other extra fields will be kept
